"""Added Version.created_with

Revision ID: 59092d41175c
Revises: 5355b569237b
Create Date: 2013-06-19 15:31:53.547392

"""

# revision identifiers, used by Alembic.
revision = '59092d41175c'
down_revision = '5355b569237b'

from alembic import op
import sqlalchemy as sa


def upgrade():
    try:
        op.add_column('Versions', sa.Column('created_with', sa.String(length=256), nullable=True))
    except sa.exc.OperationalError:
        pass

def downgrade():
    try:
        op.drop_column('Versions', 'created_with')
    except sa.exc.OperationalError:
        pass



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2014 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import unittest
from stalker import Entity, Type


class TypeTester(unittest.TestCase):
    """Tests Type class
    """

    def setUp(self):
        """set up the test
        """
        self.kwargs = {
            "name": "test type",
            'code': 'test',
            "description": "this is a test type",
            "target_entity_type": "SimpleEntity"
        }

        self.test_type = Type(**self.kwargs)

        # create another Entity with the same name of the
        # test_type for __eq__ and __ne__ tests
        self.entity1 = Entity(**self.kwargs)

    def test___auto_name__class_attribute_is_set_to_False(self):
        """testing if the __auto_name__ class attribute is set to False for
        Ticket class
        """
        self.assertFalse(Type.__auto_name__)

    def test_equality(self):
        """testing the equality operator
        """
        new_type2 = Type(**self.kwargs)

        self.kwargs["target_entity_type"] = "Asset"
        new_type3 = Type(**self.kwargs)

        self.kwargs["name"] = "a different type"
        self.kwargs["description"] = "this is a different type"
        new_type4 = Type(**self.kwargs)

        self.assertTrue(self.test_type == new_type2)
        self.assertFalse(self.test_type == new_type3)
        self.assertFalse(self.test_type == new_type4)
        self.assertFalse(self.test_type == self.entity1)

    def test_inequality(self):
        """testing the inequality operator
        """
        new_type2 = Type(**self.kwargs)

        self.kwargs["target_entity_type"] = "Asset"
        new_type3 = Type(**self.kwargs)

        self.kwargs["name"] = "a different type"
        self.kwargs["description"] = "this is a different type"
        new_type4 = Type(**self.kwargs)

        self.assertFalse(self.test_type != new_type2)
        self.assertTrue(self.test_type != new_type3)
        self.assertTrue(self.test_type != new_type4)
        self.assertTrue(self.test_type != self.entity1)

    def test_plural_class_name(self):
        """testing the plural name of Type class
        """
        self.assertTrue(self.test_type.plural_class_name, "Types")

    def test_target_entity_type_argument_can_not_be_skipped(self):
        """testing if a TypeError will be raised when the created Type doesn't
        have any target_entity_type
        """
        self.kwargs.pop("target_entity_type")
        self.assertRaises(TypeError, Type, **self.kwargs)

    def test_target_entity_type_argument_can_not_be_None(self):
        """testing if a TypeError will be raised when the target_entity_type
        argument is None
        """
        self.kwargs["target_entity_type"] = None
        self.assertRaises(TypeError, Type, **self.kwargs)

    def test_target_entity_type_argument_can_not_be_empty_string(self):
        """testing if a ValueError will be raised when the target_entity_type
        argument is an empty string
        """
        self.kwargs["target_entity_type"] = ""
        self.assertRaises(ValueError, Type, **self.kwargs)

    def test_target_entity_type_argument_accepts_strings(self):
        """testing if target_entity_type argument accepts strings
        """
        self.kwargs["target_entity_type"] = "Asset"
        # no error should be raised
        Type(**self.kwargs)

    def test_target_entity_type_argument_accepts_Python_classes(self):
        """testing if target_entity_type argument is given as a Python class
        will be converted to a string
        """
        from stalker.models.asset import Asset

        self.kwargs["target_entity_type"] = Asset
        new_type = Type(**self.kwargs)
        self.assertEqual(new_type.target_entity_type, "Asset")

    def test_target_entity_type_attribute_is_read_only(self):
        """testing if the target_entity_type attribute is read-only
        """
        self.assertRaises(AttributeError, setattr, self.test_type,
                          "target_entity_type", "Asset")

    def test_target_entity_type_attribute_is_working_properly(self):
        """testing if the target_entity_type attribute is working properly
        """
        self.assertEqual(self.test_type.target_entity_type,
                         self.kwargs["target_entity_type"])



"""create enum types

Revision ID: 46775e4a3d96
Revises: 2aeab8b376dc
Create Date: 2014-01-31 03:08:36.445876

"""

# revision identifiers, used by Alembic.

revision = '46775e4a3d96'
down_revision = '2aeab8b376dc'

from alembic import op


def upgrade():
    # rename types
    op.execute('ALTER TYPE "TaskScheduleUnit" RENAME TO "TimeUnit";')

    # create new types
    op.execute("""
        CREATE TYPE "ResourceAllocationStrategy" AS ENUM
            ('minallocated', 'maxloaded', 'minloaded', 'order', 'random');
        CREATE TYPE "TaskDependencyGapModel" AS ENUM ('length', 'duration');
        CREATE TYPE "TaskDependencyTarget" AS ENUM ('onend', 'onstart');
        CREATE TYPE "ReviewScheduleModel"
            AS ENUM ('effort', 'length', 'duration');
    """)

    # update the Task column to use the TimeUnit type instead of TaskBidUnit
    op.execute("""
        ALTER TABLE "Tasks" ALTER COLUMN bid_unit TYPE "TimeUnit"
            USING ((bid_unit::text)::"TimeUnit");
    """)

    # remove unnecessary types
    op.execute('DROP TYPE IF EXISTS "TaskBidUnit" CASCADE;')


def downgrade():
    # add necessary types
    op.execute(
        """CREATE TYPE "TaskBidUnit" AS ENUM
        ('min', 'h', 'd', 'w', 'm', 'y');
    """)

    # update the Task column to use the TimeUnit type instead of TaskBidUnit
    op.execute("""
        ALTER TABLE "Tasks" ALTER COLUMN bid_unit TYPE "TaskBidUnit"
            USING ((bid_unit::text)::"TaskBidUnit");
    """)

    # rename types
    op.execute('ALTER TYPE "TimeUnit" RENAME TO "TaskScheduleUnit";')

    # create new types
    op.execute("""
        DROP TYPE IF EXISTS "ResourceAllocationStrategy" CASCADE;
        DROP TYPE IF EXISTS "TaskDependencyGapModel" CASCADE;
        DROP TYPE IF EXISTS "TaskDependencyTarget" CASCADE;
        DROP TYPE IF EXISTS "ReviewScheduleModel" CASCADE;
    """)





# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2014 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import unittest

from sqlalchemy import Column, Integer, ForeignKey
from stalker.db.session import DBSession
from stalker.models.mixins import StatusMixin
from stalker.models.status import StatusList, Status

# create a new mixed in SimpleEntity
from stalker.models.entity import SimpleEntity


class DeclStatMixA(SimpleEntity, StatusMixin):
    __tablename__ = "DeclStatMixAs"
    __mapper_args__ = {"polymorphic_identity": "DeclStatMixA"}
    declStatMixAs_id = Column("id", Integer, ForeignKey("SimpleEntities.id"),
                              primary_key=True)

    def __init__(self, **kwargs):
        super(DeclStatMixA, self).__init__(**kwargs)
        StatusMixin.__init__(self, **kwargs)


class DeclStatMixB(SimpleEntity, StatusMixin):
    __tablename__ = "DeclStatMixBs"
    __mapper_args__ = {"polymorphic_identity": "DeclStatMixB"}
    b_id = Column("id", Integer, ForeignKey("SimpleEntities.id"),
                  primary_key=True)

    def __init__(self, **kwargs):
        super(DeclStatMixB, self).__init__(**kwargs)
        StatusMixin.__init__(self, **kwargs)


class StatusMixinTester(unittest.TestCase):
    """tests StatusMixin
    """

    def setUp(self):
        """setup the test
        """
        self.test_stat1 = Status(name="On Hold", code="OH")
        self.test_stat2 = Status(name="Work In Progress", code="WIP")
        self.test_stat3 = Status(name="Approved", code="APP")

        self.test_a_statusList = StatusList(
            name="A Statuses",
            statuses=[self.test_stat1, self.test_stat3],
            target_entity_type="DeclStatMixA",
            )

        self.test_b_statusList = StatusList(
            name="B Statuses",
            statuses=[self.test_stat2, self.test_stat3],
            target_entity_type="DeclStatMixB"
        )

        self.kwargs = {
            "name": "ozgur",
            "status_list": self.test_a_statusList
        }

    def tearDown(self):
        """clean up the test
        """
        DBSession.remove()

    def test_status_list_argument_not_set(self):
        """testing if a TypeError will be raised when the status_list argument
        is not set
        """
        self.kwargs.pop("status_list")
        self.assertRaises(TypeError, DeclStatMixA, **self.kwargs)

    def test_status_list_argument_is_not_correct(self):
        """testing if a TypeError will be raised when the given StatusList
        instance with the status_list argument is not suitable for this class
        """
        self.kwargs["status_list"] = self.test_b_statusList
        self.assertRaises(TypeError, DeclStatMixA, **self.kwargs)

    def test_status_list_working_properly(self):
        """testing if the status_list attribute is working properly
        """
        new_a_ins = DeclStatMixA(
            name="Ozgur",
            status_list=self.test_a_statusList
        )

        self.assertTrue(self.test_stat1 in new_a_ins.status_list)
        self.assertFalse(self.test_stat2 in new_a_ins.status_list)
        self.assertTrue(self.test_stat3 in new_a_ins.status_list)



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2014 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import unittest
from stalker.db import DBSession
from stalker.models.auth import User, Group


class GroupTester(unittest.TestCase):
    """tests the stalker.models.auth.Group class
    """

    @classmethod
    def setUpClass(cls):
        """sets the test in class level
        """
        DBSession.remove()

    @classmethod
    def tearDownClass(cls):
        """clear the test in class level
        """
        DBSession.remove()

    def setUp(self):
        """set up the test in method level
        """
        # create a couple of Users
        self.test_user1 = User(
            name='User1',
            login='user1',
            password='1234',
            email='user1@test.com',
        )

        self.test_user2 = User(
            name='User2',
            login='user2',
            password='1234',
            email='user1@test.com',
        )

        self.test_user3 = User(
            name='User3',
            login='user3',
            password='1234',
            email='user3@test.com',
        )

        # create a test group
        self.kwargs = {
            "name": "Test Group",
            "users": [
                self.test_user1,
                self.test_user2,
                self.test_user3
            ]
        }

        self.test_group = Group(**self.kwargs)

    def test___auto_name__class_attribute_is_set_to_False(self):
        """testing if the __auto_name__ class attribute is set to False for
        Group class
        """
        self.assertFalse(Group.__auto_name__)

    def test_users_argument_is_skipped(self):
        """testing if the users argument is skipped the users attribute will be
        an empty list
        """
        self.kwargs.pop('users')
        new_group = Group(**self.kwargs)
        self.assertEqual(new_group.users, [])

    def test_users_argument_is_not_a_list_of_User_instances(self):
        """testing if a TypeError will be raised when the users argument is not
        a list of User instances
        """
        self.kwargs['users'] = [12, 'not a user']
        self.assertRaises(TypeError, Group, **self.kwargs)

    def test_users_attribute_is_not_a_list_of_User_instances(self):
        """testing if a TypeError will be raised when the users attribute is
        not a list of User instances
        """
        self.assertRaises(
            TypeError,
            setattr,
            self.test_group,
            'users',
            [12, 'not a user']
        )

    def test_users_argument_updates_the_groups_attribute_in_the_given_User_instances(self):
        """testing if the Users given with the users argument will have the
        current Group instance in their groups attribute
        """
        self.kwargs['name'] = 'New Group'
        new_group = Group(**self.kwargs)

        for user in self.kwargs['users']:
            self.assertIn(new_group, user.groups)

    def test_users_attribute_updates_the_groups_attribute_in_the_given_User_instances(self):
        """testing if the Users given with the users attribute will have the
        current Group instance in their groups attribute
        """
        test_users = self.kwargs.pop('users')
        new_group = Group(**self.kwargs)
        new_group.users = test_users
        for user in test_users:
            self.assertIn(new_group, user.groups)



"""Added User.efficiency column

Revision ID: 59bfe820c369
Revises: af869ddfdf9
Create Date: 2014-04-26 23:50:53.880274

"""

# revision identifiers, used by Alembic.
revision = '59bfe820c369'
down_revision = 'af869ddfdf9'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('Users', sa.Column('efficiency', sa.Float(), nullable=True))
    # set default value
    op.execute('update "Users" set efficiency = 1.0')


def downgrade():
    op.drop_column('Users', 'efficiency')



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2014 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
"""Database module of Stalker.

Whenever stalker.db or something under it imported, the
:func:`stalker.db.setup` becomes available to let one setup the database.
"""

import logging

from sqlalchemy import engine_from_config
from sqlalchemy.exc import IntegrityError

from stalker import defaults
from stalker.db.declarative import Base
from stalker.db.session import DBSession
from stalker.log import logging_level

logger = logging.getLogger(__name__)
logger.setLevel(logging_level)


def setup(settings=None):
    """Utility function that helps to connect the system to the given database.

    if the database is None then the it setups using the default database in
    the settings file.

    :param settings: This is a dictionary which has keys prefixed with
        "sqlalchemy" and shows the settings. The most important one is the
        engine. The default is None, and in this case it uses the settings from
        stalker.config.Config.database_engine_settings
    """

    if settings is None:
        settings = defaults.database_engine_settings
        logger.debug('no settings given, using the default: %s' % settings)

    logger.debug("settings: %s" % settings)
    # create engine
    engine = engine_from_config(settings, 'sqlalchemy.')

    logger.debug('engine: %s' % engine)

    # create the Session class
    DBSession.remove()
    DBSession.configure(
        bind=engine,
        extension=None
    )

    # create the database
    logger.debug("creating the tables")
    Base.metadata.create_all(engine)

    # update defaults
    update_defaults_with_studio()


def update_defaults_with_studio():
    """updates the default values from Studio instance if a database and a
    Studio instance is present
    """
    if DBSession:
        with DBSession.no_autoflush:
            from stalker.models.studio import Studio
            studio = Studio.query.first()
            if studio:
                logger.debug('found a studio, updating defaults')
                studio.update_defaults()


def init():
    """fills the database with default values
    """
    logger.debug("initializing database")

    # register all Actions available for all SOM classes
    class_names = [
        'Asset', 'TimeLog', 'Client', 'Department', 'Entity',
        'FilenameTemplate', 'Group', 'ImageFormat', 'Link', 'Message', 'Note',
        'Page', 'Permission', 'Project', 'Repository', 'Review', 'Scene',
        'Sequence', 'Shot', 'SimpleEntity', 'Status', 'StatusList',
        'Structure', 'Studio', 'Tag', 'Task', 'Ticket', 'TicketLog', 'Type',
        'User', 'Vacation', 'Version']

    for class_name in class_names:
        _temp = __import__(
            'stalker',
            globals(),
            locals(),
            [class_name],
            0
        )
        class_ = eval("_temp." + class_name)
        register(class_)

    # create the admin if needed
    if defaults.auto_create_admin:
        __create_admin__()

    # create statuses
    create_ticket_statuses()
    create_entity_statuses(entity_type='Task',
                           status_names=defaults.task_status_names,
                           status_codes=defaults.task_status_codes)
    create_entity_statuses(entity_type='Asset',
                           status_names=defaults.task_status_names,
                           status_codes=defaults.task_status_codes)
    create_entity_statuses(entity_type='Shot',
                           status_names=defaults.task_status_names,
                           status_codes=defaults.task_status_codes)
    create_entity_statuses(entity_type='Sequence',
                           status_names=defaults.task_status_names,
                           status_codes=defaults.task_status_codes)
    create_entity_statuses(entity_type='Review',
                           status_names=defaults.review_status_names,
                           status_codes=defaults.review_status_codes)

    # create alembic revision table
    create_alembic_table()

    logger.debug('finished initializing the database')


def create_alembic_table():
    """creates the default alembic_version table and creates the data so that
    any new database will be considered as the latest version
    """
    # Now, this is not the correct way of doing, there is a proper way of doing
    # it and it is explained nicely in the Alembic library documentation.
    #
    # But it is simply not working when Stalker is installed as a package.
    #
    # So as a workaround here we are doing it manually
    # don't forget to update the version_num (and the corresponding test
    # whenever a new alembic revision is created)

    version_num = '5999269aad30'

    logger.debug('creating alembic_version table with version_num: %s' %
                 version_num)

    from sqlalchemy import Table, Column, Text

    table_name = 'alembic_version'

    def do_create():
        return Table(
            table_name, Base.metadata,
            Column('version_num', Text)
        )

    if table_name not in Base.metadata:
        table = do_create()
        engine = DBSession.connection().engine
        Base.metadata.create_all(engine)
        ins = table.insert().values(version_num=version_num)
        DBSession.connection().execute(ins)
        DBSession.commit()

    logger.debug('alembic_version table is created and initialized')


def __create_admin__():
    """creates the admin
    """
    from stalker.models.auth import User
    from stalker.models.department import Department

    # check if there is already an admin in the database
    admin = User.query.filter_by(name=defaults.admin_name).first()
    if admin:
        #there should be an admin user do nothing
        logger.debug("there is an admin already")
        return

    logger.debug("creating the default administrator user")

    # create the admin department
    admin_department = Department.query.filter_by(
        name=defaults.admin_department_name
    ).first()

    if not admin_department:
        admin_department = Department(name=defaults.admin_department_name)
        DBSession.add(admin_department)
        # create the admins group

    from stalker.models.auth import Group

    admins_group = Group.query \
        .filter_by(name=defaults.admin_group_name) \
        .first()

    if not admins_group:
        admins_group = Group(name=defaults.admin_group_name)
        DBSession.add(admins_group)

    # # create the admin user
    # admin = User.query \
    #     .filter_by(name=defaults.admin_name) \
    #     .first()

    # if not admin:
    admin = User(
        name=defaults.admin_name,
        login=defaults.admin_login,
        password=defaults.admin_password,
        email=defaults.admin_email,
        departments=[admin_department],
        groups=[admins_group]
    )

    admin.created_by = admin
    admin.updated_by = admin

    # update the department as created and updated by admin user
    admin_department.created_by = admin
    admin_department.updated_by = admin

    admins_group.created_by = admin
    admins_group.updated_by = admin

    DBSession.add(admin)
    DBSession.commit()


def create_ticket_statuses():
    """creates the default ticket statuses
    """
    from stalker import User

    # create as admin
    admin = User.query.filter(User.login == defaults.admin_name).first()

    # create statuses for Tickets
    ticket_names = defaults.ticket_status_names
    ticket_codes = defaults.ticket_status_codes
    create_entity_statuses('Ticket', ticket_names, ticket_codes, admin)

    # Again I hate doing this in this way
    from stalker import Type

    types = Type.query \
        .filter_by(target_entity_type="Ticket") \
        .all()
    t_names = [t.name for t in types]

    # create Ticket Types
    logger.debug("Creating Ticket Types")
    if 'Defect' not in t_names:
        ticket_type_1 = Type(
            name='Defect',
            code='Defect',
            target_entity_type='Ticket',
            created_by=admin,
            updated_by=admin
        )
        DBSession.add(ticket_type_1)

    if 'Enhancement' not in t_names:
        ticket_type_2 = Type(
            name='Enhancement',
            code='Enhancement',
            target_entity_type='Ticket',
            created_by=admin,
            updated_by=admin
        )
        DBSession.add(ticket_type_2)
    try:
        DBSession.commit()
    except IntegrityError:
        DBSession.rollback()
        logger.debug("Ticket Types are already in the database!")
    else:
        # DBSession.flush()
        logger.debug("Ticket Types are created successfully")


def create_entity_statuses(entity_type='', status_names=None,
                           status_codes=None, user=None):
    """creates the default task statuses
    """
    if not entity_type:
        raise ValueError('Please supply entity_type')

    if not status_names:
        raise ValueError('Please supply status names')

    if not status_codes:
        raise ValueError('Please supply status codes')

    # create statuses for entity
    from stalker import Status, StatusList

    logger.debug("Creating %s Statuses" % entity_type)

    statuses = Status.query.filter(Status.name.in_(status_names)).all()
    status_names_in_db = map(lambda x: x.name, statuses)

    for name, code in zip(status_names, status_codes):
        if name not in status_names_in_db:
            logger.debug('Creating Status: %s (%s)' % (name, code))
            new_status = Status(
                name=name,
                code=code,
                created_by=user,
                updated_by=user
            )
            statuses.append(new_status)
            DBSession.add(new_status)

    # create the Status List
    status_list = StatusList.query\
        .filter(StatusList.target_entity_type == entity_type)\
        .first()

    if status_list is None:
        logger.debug('No %s Status List found, creating new!' % entity_type)
        status_list = StatusList(
            name='%s Statuses' % entity_type,
            target_entity_type=entity_type,
            created_by=user,
            updated_by=user
        )
    else:
        logger.debug("%s Status List already created, updating statuses" %
                     entity_type)

    status_list.statuses = statuses
    DBSession.add(status_list)

    try:
        DBSession.commit()
    except IntegrityError as e:
        logger.debug("error in DBSession.commit, rolling back: %s" % e)
        DBSession.rollback()
    else:
        logger.debug("Created %s Statuses successfully" % entity_type)
        DBSession.flush()


def register(class_):
    """Registers the given class to the database.

    It is mainly used to create the :class:`.Action`\ s needed for the
    :class:`.User`\ s and :class:`.Group`\ s to be able to interact with the
    given class. Whatever class you have created needs to be registered.

    Example, lets say that you have a data class which is specific to your
    studio and it is not present in Stalker Object Model (SOM), so you need to
    extend SOM with a new data type. Here is a simple Data class inherited from
    the :class:`.SimpleEntity` class (which is the simplest class you should
    inherit your classes from or use more complex classes down to the
    hierarchy)::

      from sqlalchemy import Column, Integer, ForeignKey
      from stalker.models.entity import SimpleEntity

      class MyDataClass(SimpleEntity):
        '''This is an example class holding a studio specific data which is not
        present in SOM.
        '''

        __tablename__ = 'MyData'
        __mapper_arguments__ = {'polymorphic_identity': 'MyData'}

        my_data_id = Column('id', Integer, ForeignKey('SimpleEntities.c.id'),
                            primary_key=True)

    Now because Stalker is using Pyramid authorization mechanism it needs to be
    able to have an :class:`.Permission` about your new class, so you can
    assign this :class;`.Permission` to your :class:`.User`\ s or
    :class:`.Group`\ s. So you ned to register your new class with
    :func:`stalker.db.register` like shown below::

      from stalker import db
      db.register(MyDataClass)

    This will create the necessary Actions in the 'Actions' table on your
    database, then you can create :class:`.Permission`\ s and assign these to
    your :class:`.User`\ s and :class:`.Group`\ s so they are Allowed or Denied
    to do the specified Action.

    :param class_: The class itself that needs to be registered.
    """
    from stalker.models.auth import Permission

    # create the Permissions
    permissions_db = Permission.query.all()

    if not isinstance(class_, type):
        raise TypeError('To register a class please supply the class itself.')

    # register the class name to entity_types table
    from stalker import (EntityType, StatusMixin, DateRangeMixin,
                         ReferenceMixin, ScheduleMixin)

    class_name = class_.__name__

    if not EntityType.query.filter_by(name=class_name).first():
        new_entity_type = EntityType(class_name)
        # update attributes
        if issubclass(class_, StatusMixin):
            new_entity_type.statusable = True
        if issubclass(class_, DateRangeMixin):
            new_entity_type.dateable = True
        if issubclass(class_, ScheduleMixin):
            new_entity_type.schedulable = True
        if issubclass(class_, ReferenceMixin):
            new_entity_type.accepts_references = True

        DBSession.add(new_entity_type)

    for action in defaults.actions:
        for access in ['Allow', 'Deny']:
            permission_obj = Permission(access, action, class_name)
            if permission_obj not in permissions_db:
                DBSession.add(permission_obj)

    try:
        DBSession.commit()
    except IntegrityError:
        DBSession.rollback()
    # else:
    #     DBSession.flush()



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2014 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import unittest

from sqlalchemy import Column, Integer

from stalker import db
from stalker.db.session import DBSession
from stalker.models.mixins import ACLMixin
from stalker.db.declarative import Base
from stalker.models.auth import Permission


# create a new class and mix it with ACLMixin
class TestClassForACL(Base, ACLMixin):
    __tablename__ = 'TestClassForACLs'
    id = Column(Integer, primary_key=True)

    def __init__(self):
        self.name = None


class ACLMixinTester(unittest.TestCase):
    """tests the stalker.models.mixins.ACLMixin class
    """

    @classmethod
    def setUpClass(cls):
        """setup the test in class level
        """
        DBSession.remove()
        DBSession.configure()

    @classmethod
    def tearDownClass(cls):
        """cleanup the test in class level
        """
        DBSession.remove()
        DBSession.configure()

    def setUp(self):
        """setup the test
        """
        db.setup()

        # create permissions
        self.test_perm1 = Permission(
            access='Allow',
            action='Create',
            class_name='Something'
        )
        self.test_instance = TestClassForACL()
        self.test_instance.name = 'Test'
        self.test_instance.permissions.append(self.test_perm1)

    def tearDown(self):
        """clean the test
        """
        DBSession.remove()

    def test_permission_attribute_accept_Permission_instances_only(self):
        """testing if the permissions attribute accepts only Permission
        instances
        """
        self.assertRaises(TypeError, setattr, self.test_instance, [234])

    def test_permission_attribute_is_working_properly(self):
        """testing if the permissions attribute is working properly
        """
        self.assertEqual(self.test_instance.permissions, [self.test_perm1])

    def test_acl_property_returns_a_list(self):
        """testing if the __acl__ property returns a list
        """
        self.assertTrue(isinstance(self.test_instance.__acl__, list))

    def test_acl_property_returns_a_proper_ACL_list(self):
        """testing if the __acl__ property returns a proper ACL list according
        to the given permissions
        """
        self.assertEqual(
            self.test_instance.__acl__,
            [('Allow', 'TestClassForACL:Test', 'Create_Something')]
        )



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2014 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from sqlalchemy import Column, Integer, ForeignKey, Text
from sqlalchemy.orm import validates
from stalker import ProjectMixin, Entity


class Page(Entity, ProjectMixin):
    """A simple Wiki page implementation.

    Wiki in Stalker are managed per Project. That is, all Wiki pages are
    related to a Project.

    Stalker wiki pages are very simple in terms of data it holds. It has only
    one :attr:`.title` and one :attr:`.content` an some usual audit info coming
    from :class:`.SimpleEntity` and a :attr:`.project` coming from
    :class:`.ProjectMixin`.

    :param str title: The title of this Page
    :param str content: The content of this page. Can contain any kind of
      string literals including HTML tags etc.
    """

    __auto_name__ = True
    __tablename__ = 'Pages'
    __mapper_args__ = {'polymorphic_identity': 'Page'}
    page_id = Column('id', Integer, ForeignKey('Entities.id'),
                     primary_key=True)

    title = Column(Text)
    content = Column(Text)

    def __init__(self, title='', content='', project=None, **kwargs):
        kwargs['project'] = project
        super(Page, self).__init__(**kwargs)
        ProjectMixin.__init__(self, **kwargs)

        self.title = title
        self.content = content

    @validates('title')
    def _validate_title(self, key, title):
        """validates the given title value
        """
        if not isinstance(title, str):
            raise TypeError(
                '%(class)s.title should be a string, not %(title_class)s' %
                {
                    'class': self.__class__.__name__,
                    'title_class': title.__class__.__name__
                }
            )

        if not title:
            raise ValueError(
                '%(class)s.title can not be empty' %
                {
                    'class': self.__class__.__name__
                }
            )

        return title

    @validates('content')
    def _validate_content(self, key, content):
        """validates the given content value
        """
        if content is None:
            content = ''

        if not isinstance(content, str):
            raise TypeError(
                '%(class)s.content should be a string, not %(content_class)s' %
                {
                    'class': self.__class__.__name__,
                    'content_class': content.__class__.__name__
                }
            )

        return content




from datetime import datetime
from stalker import db, User, Department, Project, ImageFormat
from stalker.db.session import session

db.setup("sqlite:////tmp/studio.db")

myUser = User(
    name="Erkan Ozgur Yilmaz",
    login_name="eoyilmaz",
    email="eoyilmaz@gmail.com",
    password="secret",
    description="This is me"
)

tds_department = Department(
    name="TDs",
    description="This is the TDs department"
)

tds_department.users.append(myUser)

all_departments = Department.query.all()
all_members_of_dep = all_departments[0].members
print all_members_of_dep[0].first_name

session.add(myUser)
session.add(tds_department)

new_project = Project(name="Fancy Commercial")

new_project.description = """The commercial is about this fancy product. The
                             client want us to have a shinny look with their
                             product bla bla bla..."""
new_project.image_format = ImageFormat(name="HD 1080", width=1920, height=1080)
new_project.fps = 25
new_project.due = datetime(2011,2,15)
new_project.lead = myUser

from stalker import Type

commercial_project_type = Type(name="Commercial")
new_project.type = commercial_project_type

session.add(new_project)
session.commit()


from stalker import Sequence
seq1 = Sequence(name="Sequence 1", code="SEQ1", project=new_project)

from stalker import Shot

sh001 = Shot(name="Shot 1", code="SH001", project=new_project)
sh002 = Shot(name="Shot 2", code="SH002", project=new_project)
sh003 = Shot(name="Shot 3", code="SH003", project=new_project)

# assign them to the sequence
seq1.shots.extend([sh001, sh002, sh003])




# the order of the PipelineSteps are not important
shot_pSteps = [previs, match, anim, layout, light, comp]

# create the asset type
shot_asset_type = AssetType(name="Shot", steps=shot_pSteps)

# and set our shot objects asset_type to shot_asset_type
# 
# instead of writing down shot1.type = shot_asset_type
# we are going to do something more interesting
# (eventhough it is inefficient)

for shot in seq1.shots:
    shot.type = shot_asset_type


from datetime import timedelta
from stalker.models import Task

previs_task = Task(
                  name="Previs",
                  resources=[myUser],
                  bid=timedelta(days=1),
                  pipeline_step=previs
              )

mm_task     = Task(
                  name="Match Move",
                  resources=[myUser],
                  bid=timedelta(days=2),
                  pipeline_step=matchmove
              )

anim_task   = Task(
                  name="Animation",
                  resources=[myUser],
                  bid=timedelta(days=2),
                  pipeline_step=anim
              )

layout_task = Task(
                  name="Layout",
                  resources=[myUser],
                  bid=timdelta(hours=2),
                  pipeline_step=layout
              )

light_task  = Task(
                  name="Lighting",
                  resources=[myUser],
                  bid=timedelta(days=2),
                  pipeline_step=light
              )

comp_task   = Task(
                  name="Compositing",
                  resources=[myUser],
                  bid=timedelta(days=2),
                  pipeline_step=comp
              )

sh001.tasks = [previs_task,
               mm_task,
               anim_task,
               layout_task,
               light_task,
               comp_task]

# animation needs match moving and previs to be finished
anim_task.depends = [mm_task, previs_task]

# compositing can not start before anything rendered or animated
comp_task.depends = [light_task, anim_task]

# lighting can not be done before the layout is finished
light_task.depends = [layout_task]


session.commit()


from stalker.models import Repository
repo1 = Repository(
    name="Commercial Repository",
    description="""This is where the commercial projects are going to be
                   stored"""
)

repo1.windows_path = "M:\\PROJECTS"
repo1.linux_path   = "/mnt/M"
repo1.osx_path     = "/Volumes/M"

print repo1.path
# outputs:
# if you are running the command on a computer with Windows it will output:
#
# M:\PROJECTS
# 
# and for Linux:
# /mnt/M 
# 
# for OSX:
# /Volumes/M
#

from stalker.models import Structure

structure1 = Structure(
    name="Commercial Projects Structure",
    description="""This is a project structure, which can be used for simple
        commercial projects"""
)

# lets create the folder structure as a Jinja2 template
project_template = """
   {{ project.code }}
   {{ project.code }}/Assets
   {{ project.code }}/Sequences"
   
   {% if project.sequences %}
       {% for sequence in project.sequences %}
           {% set seq_path = project.code + '/Sequences/' + sequence.code %}
           {{ seq_path }}
           {{ seq_path }}/Edit
           {{ seq_path }}/Edit/AnimaticStoryboard
           {{ seq_path }}/Edit/Export
           {{ seq_path }}/Storyboard
           {{ seq_path }}/Shots
           
           {% if sequence.shots %}
               {% for shot in sequence.shots %}
                   {% set shot_path = seq_path + '/SHOTS/' + shot.code %}
                   {{ shot_path }}
               {% endfor %}
           {% endif %}
           
       {% endfor %}
   
   {% endif %}
   
   {{ project.code }}/References
"""

structure1.project_template = project_template


#M:/PROJECTS/FANCY_COMMERCIAL
#M:/PROJECTS/FANCY_COMMERCIAL/Assets
#M:/PROJECTS/FANCY_COMMERCIAL/References
#M:/PROJECTS/FANCY_COMMERCIAL/Sequences
#M:/PROJECTS/FANCY_COMMERCIAL/Sequences/SEQ1
#M:/PROJECTS/FANCY_COMMERCIAL/Sequences/SEQ1/Edit
#M:/PROJECTS/FANCY_COMMERCIAL/Sequences/SEQ1/Edit/AnimaticStoryboard
#M:/PROJECTS/FANCY_COMMERCIAL/Sequences/SEQ1/Edit/Export
#M:/PROJECTS/FANCY_COMMERCIAL/Sequences/SEQ1/Storyboard
#M:/PROJECTS/FANCY_COMMERCIAL/Sequences/SEQ1/Shots
#M:/PROJECTS/FANCY_COMMERCIAL/Sequences/SEQ1/Shots/SH001
#M:/PROJECTS/FANCY_COMMERCIAL/Sequences/SEQ1/Shots/SH002
#M:/PROJECTS/FANCY_COMMERCIAL/Sequences/SEQ1/Shots/SH003




# -*- coding: utf-8 -*-
# stalker_pyramid
# Copyright (C) 2013 Erkan Ozgur Yilmaz
#
# This file is part of stalker_pyramid.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

import unittest
from sqlalchemy import Column, Integer, ForeignKey
from stalker import SimpleEntity, ScheduleMixin, defaults


class MixedInClass(SimpleEntity, ScheduleMixin):
    """A simple class derived from SimpleEntity and mixed in with ScheduleMixin
    used for tests
    """

    __tablename__ = "ScheduleMixFooMixedInClasses"
    __mapper_args__ = {"polymorphic_identity": "ScheduleMixFooMixedInClass"}
    schedMixFooMixedInClass_id = Column(
        "id", Integer,
        ForeignKey("SimpleEntities.id"),
        primary_key=True
    )

    def __init__(self, **kwargs):
        SimpleEntity.__init__(self, **kwargs)
        ScheduleMixin.__init__(self, **kwargs)


class ScheduleMixinTestCase(unittest.TestCase):
    """tests the ScheduleMixin
    """

    def setUp(self):
        """set up the test
        """
        self.kwargs = {
            'name': 'Test Object',
            'schedule_timing': 1,
            'schedule_unit': 'h',
            'schedule_model': 'effort',
            'schedule_constraint': 0
        }
        self.test_obj = MixedInClass(**self.kwargs)

    def tearDown(self):
        from stalker.db.session import DBSession
        DBSession.remove()

    def test_schedule_model_attribute_is_effort_by_default(self):
        """testing if the schedule_model is effort by default
        """
        self.assertEqual(self.test_obj.schedule_model, 'effort')

    def test_schedule_model_argument_is_None(self):
        """testing if the schedule model attribute will be 'effort' if the
        schedule_model argument is set to None
        """
        self.kwargs['schedule_model'] = None
        new_task = MixedInClass(**self.kwargs)
        self.assertEqual(
            new_task.schedule_model,
            'effort'
        )

    def test_schedule_model_attribute_is_set_to_None(self):
        """testing if the schedule_model will be 'effort' if it is set to None
        """
        self.test_obj.schedule_model = None
        self.assertEqual(
            self.test_obj.schedule_model,
            'effort'
        )

    def test_schedule_model_argument_is_not_a_string(self):
        """testing if a TypeError will be raised when the schedule_model
        argument is not a string
        """
        self.kwargs['schedule_model'] = 234
        self.assertRaises(TypeError, MixedInClass, **self.kwargs)

    def test_schedule_model_attribute_is_not_a_string(self):
        """testing if a TypeError will be raised when the schedule_model
        attribute is set to a value other than a string
        """
        self.assertRaises(
            TypeError, setattr, self.test_obj, 'schedule_model', 2343
        )

    def test_schedule_model_argument_is_not_in_correct_value(self):
        """testing if a ValueError will be raised when the schedule_model
        argument is not in correct value
        """
        self.kwargs['schedule_model'] = 'not in the list'
        self.assertRaises(ValueError, MixedInClass, **self.kwargs)

    def test_schedule_model_attribute_is_not_in_correct_value(self):
        """testing if a ValueError will be raised when the schedule_model
        attribute is not set to a correct value
        """
        self.assertRaises(
            ValueError, setattr, self.test_obj, 'schedule_model',
            'not in the list'
        )

    def test_schedule_model_argument_is_working_properly(self):
        """testing if the schedule_model argument value is correctly passed to
        the schedule_model attribute
        """
        test_value = 'duration'
        self.kwargs['schedule_model'] = test_value
        new_task = MixedInClass(**self.kwargs)
        self.assertEqual(new_task.schedule_model, test_value)

    def test_schedule_model_attribute_is_working_properly(self):
        """testing if the schedule_model attribute is working properly
        """
        test_value = 'duration'
        self.assertNotEqual(
            self.test_obj.schedule_model, test_value
        )
        self.test_obj.schedule_model = test_value
        self.assertEqual(
            self.test_obj.schedule_model, test_value
        )

    def test_schedule_constraint_is_0_by_default(self):
        """testing if the schedule_constraint attribute is None by default
        """
        self.assertEqual(self.test_obj.schedule_constraint, 0)

    def test_schedule_constraint_argument_is_skipped(self):
        """testing if the schedule_constraint attribute will be 0 if
        schedule_constraint is skipped
        """
        try:
            self.kwargs.pop('schedule_constraint')
        except KeyError:
            pass
        new_task = MixedInClass(**self.kwargs)
        self.assertEqual(new_task.schedule_constraint, 0)

    def test_schedule_constraint_argument_is_None(self):
        """testing if the schedule_constraint attribute will be 0 if
        schedule_constraint is None
        """
        self.kwargs['schedule_constraint'] = None
        new_task = MixedInClass(**self.kwargs)
        self.assertEqual(new_task.schedule_constraint, 0)

    def test_schedule_constraint_attribute_is_set_to_None(self):
        """testing if the schedule_constraint attribute will be 0 if
        it is set to None
        """
        self.test_obj.schedule_constraint = None
        self.assertEqual(self.test_obj.schedule_constraint, 0)

    def test_schedule_constraint_argument_is_not_an_integer(self):
        """testing if a TypeError will be raised when the schedule_constraint
        argument is not an integer
        """
        self.kwargs['schedule_constraint'] = 'not an integer'
        self.assertRaises(TypeError, MixedInClass, **self.kwargs)

    def test_schedule_constraint_attribute_is_not_an_integer(self):
        """testing if a TypeError will be raised when the schedule_constraint
        attribute is set to a value other than an integer
        """
        self.assertRaises(TypeError, setattr, self.test_obj,
                          'schedule_constraint', 'not an integer')

    def test_schedule_constraint_argument_is_working_properly(self):
        """testing if the schedule_constraint argument value is correctly
        passed to schedule_constraint attribute
        """
        test_value = 2
        self.kwargs['schedule_constraint'] = test_value
        new_task = MixedInClass(**self.kwargs)
        self.assertEqual(new_task.schedule_constraint, test_value)

    def test_schedule_constraint_attribute_is_working_properly(self):
        """testing if the schedule_constraint attribute value is correctly
        changed
        """
        test_value = 3
        self.test_obj.schedule_constraint = test_value
        self.assertEqual(self.test_obj.schedule_constraint, test_value)

    def test_schedule_constraint_argument_value_is_out_of_range(self):
        """testing if the value of schedule_constraint argument value will be
        clamped to the [0-3] range if it is out of range
        """
        self.kwargs['schedule_constraint'] = -1
        new_task = MixedInClass(**self.kwargs)
        self.assertEqual(new_task.schedule_constraint, 0)

        self.kwargs['schedule_constraint'] = 4
        new_task = MixedInClass(**self.kwargs)
        self.assertEqual(new_task.schedule_constraint, 3)

    def test_schedule_constraint_attribute_value_is_out_of_range(self):
        """testing if the value of schedule_constraint attribute value will be
        clamped to the [0-3] range if it is out of range
        """
        self.test_obj.schedule_constraint = -1
        self.assertEqual(self.test_obj.schedule_constraint, 0)

        self.test_obj.schedule_constraint = 4
        self.assertEqual(self.test_obj.schedule_constraint, 3)

    def test_schedule_timing_argument_skipped(self):
        """testing if the schedule_timing attribute will be equal to 1 hour if
        the schedule_timing argument is skipped
        """
        self.kwargs.pop("schedule_timing")
        new_task = MixedInClass(**self.kwargs)

        self.assertEqual(new_task.schedule_timing,
                         MixedInClass.__default_schedule_timing__)

    def test_schedule_timing_argument_is_None(self):
        """testing if the schedule_timing attribute will be equal to the
        stalker.config.Config.timing_resolutio.seconds/3600 if the
        schedule_timing argument is None
        """
        self.kwargs["schedule_timing"] = None
        new_task = MixedInClass(**self.kwargs)
        self.assertEqual(
            new_task.schedule_timing,
            defaults.timing_resolution.seconds / 60.0
        )

    def test_schedule_timing_attribute_is_set_to_None(self):
        """testing if the schedule_timing attribute will be equal to the
        stalker.config.Config.timing_resolution.seconds/3600 if it is set to
        None
        """
        self.test_obj.schedule_timing = None
        self.assertEqual(
            self.test_obj.schedule_timing,
            defaults.timing_resolution.seconds / 60.0
        )

    def test_schedule_timing_argument_is_not_an_integer_or_float(self):
        """testing if a TypeError will be raised when the schedule_timing
        is not an integer or float
        """
        self.kwargs["schedule_timing"] = '10d'
        self.assertRaises(TypeError, MixedInClass, **self.kwargs)

    def test_schedule_timing_attribute_is_not_an_integer_or_float(self):
        """testing if a TypeError will be raised when the schedule_timing
        attribute is not set to an integer or float
        """
        self.assertRaises(TypeError, setattr, self.test_obj,
                          'schedule_timing', '10d')

    def test_schedule_timing_attribute_is_working_properly(self):
        """testing if the schedule_timing attribute is working properly
        """
        test_value = 18
        self.test_obj.schedule_timing = test_value
        self.assertEqual(self.test_obj.schedule_timing, test_value)

    def test_schedule_unit_argument_skipped(self):
        """testing if the schedule_unit attribute will use the default value if
        the schedule_unit argument is skipped
        """
        self.kwargs.pop("schedule_unit")
        new_task = MixedInClass(**self.kwargs)
        self.assertEqual(new_task.schedule_unit,
                         MixedInClass.__default_schedule_unit__)

    def test_schedule_unit_argument_is_None(self):
        """testing if the schedule_unit attribute will use the default value if
        the schedule_unit argument is None
        """
        self.kwargs["schedule_unit"] = None
        new_task = MixedInClass(**self.kwargs)
        self.assertEqual(new_task.schedule_unit,
                         MixedInClass.__default_schedule_unit__)

    def test_schedule_unit_attribute_is_set_to_None(self):
        """testing if the schedule_unit attribute will use the default value if
        it is set to None
        """
        self.test_obj.schedule_unit = None
        self.assertEqual(self.test_obj.schedule_unit,
                         MixedInClass.__default_schedule_unit__)

    def test_schedule_unit_argument_is_not_a_string(self):
        """testing if a TypeError will be raised when the schedule_unit is not
        an integer
        """
        self.kwargs["schedule_unit"] = 10
        self.assertRaises(TypeError, MixedInClass, **self.kwargs)

    def test_schedule_unit_attribute_is_not_a_string(self):
        """testing if a TypeError will be raised when the schedule_unit
        attribute is not set to a string
        """
        self.assertRaises(TypeError, setattr, self.test_obj, 'schedule_unit',
                          23)

    def test_schedule_unit_attribute_is_working_properly(self):
        """testing if the schedule_unit attribute is working properly
        """
        test_value = 'w'
        self.test_obj.schedule_unit = test_value
        self.assertEqual(self.test_obj.schedule_unit, test_value)

    def test_schedule_unit_argument_value_is_not_in_defaults_datetime_units(self):
        """testing if a ValueError will be raised when the schedule_unit value
        is not in stalker.config.Config.datetime_units list
        """
        self.kwargs['schedule_unit'] = 'os'
        self.assertRaises(ValueError, MixedInClass, **self.kwargs)

    def test_schedule_unit_attribute_value_is_not_in_defaults_datetime_units(self):
        """testing if a ValueError will be raised when it is set to a value
        which is not in stalker.config.Config.datetime_units list
        """
        self.assertRaises(ValueError, setattr, self.test_obj, 'schedule_unit',
                          'so')

    def test_least_meaningful_time_unit_is_working_properly(self):
        """testing if the least_meaningful_time_unit is working properly
        """
        defaults.daily_working_hours = 9
        defaults.weekly_working_days = 5
        defaults.weekly_working_hours = 45
        defaults.yearly_working_days = 52.1428 * 5

        test_values = [
            [[60, True], (1, 'min')],
            [[125, True], (2, 'min')],
            [[1800, True], (30, 'min')],
            [[3600, True], (1, 'h')],
            [[5400, True], (90, 'min')],
            [[6000, True], (100, 'min')],
            [[7200, True], (2, 'h')],
            [[9600, True], (160, 'min')],
            [[10000, True], (166, 'min')],
            [[12000, True], (200, 'min')],
            [[14400, True], (4, 'h')],
            [[15000, True], (250, 'min')],
            [[18000, True], (5, 'h')],
            [[32400, True], (1, 'd')],
            [[32400, False], (9, 'h')],
            [[64800, True], (2, 'd')],
            [[64800, False], (18, 'h')],
            [[86400, True], (24, 'h')],
            [[86400, False], (1, 'd')],
            [[162000, True], (1, 'w')],
            [[162000, False], (45, 'h')],
            [[648000, True], (1, 'm')],
            [[648000, False], (180, 'h')],
            [[8424000, True], (1, 'y')],
            [[8424000, False], (2340, 'h')]
        ]

        for test_value in test_values:
            input_value = test_value[0]
            expected_result = test_value[1]
            self.assertEqual(
                expected_result,
                self.test_obj.least_meaningful_time_unit(*input_value)
            )

    def test_to_seconds_is_working_properly(self):
        """testing if the to_seconds method is working properly
        """
        defaults.daily_working_hours = 9
        defaults.weekly_working_days = 5
        defaults.weekly_working_hours = 45
        defaults.yearly_working_days = 52.1428 * 5

        test_values = [
            # effort values
            ['effort', 1, 'min', 60],
            ['effort', 1, 'h', 3600],
            ['effort', 1, 'd', 32400],
            ['effort', 1, 'w', 162000],
            ['effort', 1, 'm', 648000],
            ['effort', 1, 'y', 8424000],

            # lenth values
            ['length', 1, 'min', 60],
            ['length', 1, 'h', 3600],
            ['length', 1, 'd', 32400],
            ['length', 1, 'w', 162000],
            ['length', 1, 'm', 648000],
            ['length', 1, 'y', 8424000],

            # duration values
            ['duration', 1, 'min', 60],
            ['duration', 1, 'h', 3600],
            ['duration', 1, 'd', 86400],
            ['duration', 1, 'w', 604800],
            ['duration', 1, 'm', 2419200],
            ['duration', 1, 'y', 31536000]
        ]

        for test_value in test_values:
            self.test_obj.schedule_model = test_value[0]
            self.test_obj.schedule_timing = test_value[1]
            self.test_obj.schedule_unit = test_value[2]
            self.assertEqual(
                test_value[3],
                self.test_obj.to_seconds(
                    self.test_obj.schedule_timing,
                    self.test_obj.schedule_unit,
                    self.test_obj.schedule_model
                )
            )

    def test_schedule_seconds_is_working_properly(self):
        """testing if the schedule_seconds property is working properly
        """
        defaults.daily_working_hours = 9
        defaults.weekly_working_days = 5
        defaults.weekly_working_hours = 45
        defaults.yearly_working_days = 52.1428 * 5

        test_values = [
            # effort values
            ['effort', 1, 'min', 60],
            ['effort', 1, 'h', 3600],
            ['effort', 1, 'd', 32400],
            ['effort', 1, 'w', 162000],
            ['effort', 1, 'm', 648000],
            ['effort', 1, 'y', 8424000],

            # lenth values
            ['length', 1, 'min', 60],
            ['length', 1, 'h', 3600],
            ['length', 1, 'd', 32400],
            ['length', 1, 'w', 162000],
            ['length', 1, 'm', 648000],
            ['length', 1, 'y', 8424000],

            # duration values
            ['duration', 1, 'min', 60],
            ['duration', 1, 'h', 3600],
            ['duration', 1, 'd', 86400],
            ['duration', 1, 'w', 604800],
            ['duration', 1, 'm', 2419200],
            ['duration', 1, 'y', 31536000]
        ]

        for test_value in test_values:
            self.test_obj.schedule_model = test_value[0]
            self.test_obj.schedule_timing = test_value[1]
            self.test_obj.schedule_unit = test_value[2]
            self.assertEqual(
                test_value[3],
                self.test_obj.schedule_seconds
            )

    # def test_schedule_timing_and_schedule_unit_are_converted_to_the_least_meaningful_unit(self):
    #     """testing if the schedule_unit is converted to the least meaningful
    #     unit automatically
    #     """
    #     defaults.daily_working_hours = 9
    #     self.kwargs['schedule_timing'] = 9
    #     self.kwargs['schedule_unit'] = 'h'
    # 
    #     test_obj = MixedInClass(**self.kwargs)
    #     self.assertEqual(
    #         test_obj.schedule_timing, 1
    #     )
    #     self.assertEqual(
    #         test_obj.schedule_unit, 'd'
    #     )



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2014 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import copy
import logging
import time
import datetime
from math import ceil

from sqlalchemy import (Column, Integer, ForeignKey, Interval, Boolean,
                        DateTime, PickleType)
from sqlalchemy.orm import validates, relationship, synonym, reconstructor

from stalker import db, defaults, log
from stalker.models.entity import SimpleEntity, Entity
from stalker.models.mixins import DateRangeMixin, WorkingHoursMixin
from stalker.models.schedulers import SchedulerBase

logger = logging.getLogger(__name__)
logger.setLevel(log.logging_level)


class Studio(Entity, DateRangeMixin, WorkingHoursMixin):
    """Manage all the studio information at once.

    With Stalker you can manage all you Studio data by using this class. Studio
    knows all the projects, all the departments, all the users and every thing
    about the studio. But the most important part of the Studio is that it can
    schedule all the Projects by using TaskJuggler.

    Studio class is kind of the database itself::

      studio = Studio()

      # simple data
      studio.projects
      studio.active_projects
      studio.inactive_projects
      studio.departments
      studio.users

      # project management
      studio.to_tjp          # a tjp representation of the studio with all
                             # its projects, departments and resources etc.

      studio.schedule() # schedules all the active projects at once

    **Working Hours**

    In Stalker, Studio class also manages the working hours of the studio.
    Allowing project tasks to be scheduled to be scheduled in those hours.

    **Vacations**

    Studio wide vacations are managed by the Studio class.

    **Scheduling**

    .. versionadded: 0.2.5
       Schedule Info Attributes

    There are a couple of attributes those become pretty interesting when used
    together with the Studio instance while using the scheduling part of the
    Studio. Please refer to the attribute documentation for each attribute:

      :attr:`.is_scheduling`
      :attr:`.last_scheduled_at`
      :attr:`.last_scheduled_by`
      :attr:`.last_schedule_message`

    :param int daily_working_hours: An integer specifying the daily working
      hours for the studio. It is another critical value attribute which
      TaskJuggler uses mainly converting working day values to working hours
      (1d = 10h kind of thing).

    :param now: The now attribute overrides the TaskJugglers ``now`` attribute
      allowing the user to schedule the projects as if the scheduling is done
      on that date. The default value is the rounded value of
      datetime.datetime.now().

    :type now: datetime.datetime

    :param timing_resolution: The timing_resolution of the datetime.datetime
      object in datetime.timedelta. Uses ``timing_resolution`` settings in the
      :class:`stalker.config.Config` class which defaults to 1 hour. Setting
      the timing_resolution to less then 5 minutes is not suggested because it
      is a limit for TaskJuggler.

    :type timing_resolution: datetime.timedelta

    """
    __auto_name__ = False
    __tablename__ = 'Studios'
    __mapper_args__ = {'polymorphic_identity': 'Studio'}

    studio_id = Column(
        'id',
        Integer,
        ForeignKey('Entities.id'),
        primary_key=True,
    )

    _timing_resolution = Column("timing_resolution", Interval)

    is_scheduling = Column(Boolean, default=False)
    is_scheduling_by_id = Column(
        Integer,
        ForeignKey('Users.id'),
        doc='The id of the user who is scheduling the Studio projects right '
            'now'
    )
    is_scheduling_by = relationship(
        'User',
        primaryjoin='Studios.c.is_scheduling_by_id==Users.c.id',
        doc='The User who is scheduling the Studio projects right now'
    )
    scheduling_started_at = Column(
        DateTime,
        doc='Stores when the current scheduling is started at, it is a good '
            'measure for measuring if the last schedule is not correctly '
            'finished'
    )
    last_scheduled_at = Column(
        DateTime,
        doc='Stores the last schedule date'
    )
    last_scheduled_by_id = Column(
        Integer,
        ForeignKey('Users.id'),
        doc='The id of the user who has last scheduled the Studio projects'
    )
    last_scheduled_by = relationship(
        'User',
        primaryjoin='Studios.c.last_scheduled_by_id==Users.c.id',
        doc='The User who has last scheduled the Studio projects'
    )
    last_schedule_message = Column(
        PickleType,
        doc='Holds the last schedule message, generally coming generated by '
        'TaskJuggler'
    )

    def __init__(self,
                 daily_working_hours=None,
                 now=None,
                 timing_resolution=None,
                 **kwargs):
        super(Studio, self).__init__(**kwargs)
        DateRangeMixin.__init__(self, **kwargs)
        WorkingHoursMixin.__init__(self, **kwargs)
        self.timing_resolution = timing_resolution
        self.daily_working_hours = daily_working_hours
        self._now = None
        self.now = self._validate_now(now)
        self._scheduler = None

        # update defaults
        self.update_defaults()

    @property
    def daily_working_hours(self):
        """a shortcut for Studio.working_hours.daily_working_hours
        """
        return self.working_hours.daily_working_hours

    @daily_working_hours.setter
    def daily_working_hours(self, dwh):
        """a shortcut for Studio.working_hours.daily_working_hours
        """
        self.working_hours.daily_working_hours = dwh

    def update_defaults(self):
        """updates the default values with the studio
        """
        # TODO: add update_defaults() to attribute edit/update methods,
        #       so we will always have an up to date info about the working
        #       hours.

        logger.debug('updating defaults with Studio instance')
        from stalker import defaults
        try:
            if self.daily_working_hours:
                defaults.daily_working_hours = self.daily_working_hours
                logger.debug(
                    'updated defaults.daily_working_hours: %s' %
                    defaults.daily_working_hours
                )
            else:
                logger.debug('can not update defaults.daily_working_hours')
        except AttributeError:
            # The Studio and WorkingHours classes has changed from
            # v0.2.3 to v0.2.5 and with this change it is simply
            # no possible to initialize the db if we insist to update the
            # defaults for daily_working_hours, because there is no
            # WorkingHours._daily_working_hours in versions before than
            # v0.2.5, so just skip it for at least the studio instance
            # in the database has been updated.
            logger.debug(
                'Can not update defaults.daily_working_hours, WorkingHours '
                'version mismatch'
            )

        if self.weekly_working_days:
            defaults.weekly_working_days = self.weekly_working_days
            logger.debug(
                'updated defaults.weekly_working_days: %s' %
                defaults.weekly_working_days
            )
        else:
            logger.debug('can not update defaults.weekly_working_days')

        if self.weekly_working_hours:
            defaults.weekly_working_hours = self.weekly_working_hours
            logger.debug(
                'updated defaults.weekly_working_hours: %s' %
                defaults.weekly_working_hours
            )
        else:
            logger.debug('can not update defaults.weekly_working_hours')

        if self.yearly_working_days:
            defaults.yearly_working_days = self.yearly_working_days
            logger.debug(
                'updated defaults.yearly_working_days: %s' %
                defaults.yearly_working_days
            )
        else:
            logger.debug('can not update defaults.yearly_working_days')

        if self.timing_resolution:
            defaults.timing_resolution = self.timing_resolution
            logger.debug(
                'updated defaults.timing_resolution: %s' %
                defaults.timing_resolution
            )
        else:
            logger.debug('can not update defaults.timing_resolution')

        logger.debug("""done updating defaults:
        daily_working_hours  : %(daily_working_hours)s
        weekly_working_days  : %(weekly_working_days)s
        weekly_working_hours : %(weekly_working_hours)s
        yearly_working_days  : %(yearly_working_days)s
        timing_resolution    : %(timing_resolution)s
        """ % {
            'daily_working_hours': defaults.daily_working_hours,
            'weekly_working_days': defaults.weekly_working_days,
            'weekly_working_hours': defaults.weekly_working_hours,
            'yearly_working_days': defaults.yearly_working_days,
            'timing_resolution': defaults.timing_resolution,
        })

    @reconstructor
    def __init_on_load__(self):
        """update defaults on load
        """
        self.update_defaults()

    def _validate_now(self, now_in):
        """validates the given now_in value
        """
        if now_in is None:
            now_in = datetime.datetime.now()

        if not isinstance(now_in, datetime.datetime):
            raise TypeError(
                '%s.now attribute should be an instance of datetime.datetime, '
                'not %s' %
                (self.__class__.__name__, now_in.__class__.__name__)
            )

        return self.round_time(now_in)

    @property
    def now(self):
        """now getter
        """
        try:
            if self._now is None:
                self._now = self.round_time(datetime.datetime.now())
        except AttributeError:
            setattr(self, '_now', self.round_time(datetime.datetime.now()))
        return self._now

    @now.setter
    def now(self, now_in):
        """now setter
        """
        self._now = self._validate_now(now_in)

    def _validate_scheduler(self, scheduler_in):
        """validates the given scheduler_in value
        """
        if scheduler_in is not None:
            if not isinstance(scheduler_in, SchedulerBase):
                raise TypeError(
                    '%s.scheduler should be an instance of '
                    'stalker.models.scheduler.SchedulerBase, not %s' %
                    (self.__class__.__name__, scheduler_in.__class__.__name__)
                )
        return scheduler_in

    @property
    def scheduler(self):
        """scheduler getter
        """
        return self._scheduler

    @scheduler.setter
    def scheduler(self, scheduler_in):
        """the scheduler setter
        """
        self._scheduler = self._validate_scheduler(scheduler_in)

    @property
    def to_tjp(self):
        """converts the studio to a tjp representation
        """
        from jinja2 import Template

        temp = Template(
            defaults.tjp_studio_template,
            trim_blocks=True,
            lstrip_blocks=True
        )
        start = time.time()
        rendered_template = temp.render({
            'studio': self,
            'datetime': datetime,
            'now': self.round_time(self.now).strftime('%Y-%m-%d-%H:%M')
        })
        end = time.time()
        logger.debug('render studio to tjp took: %s seconds' % (end - start))
        return rendered_template

    @property
    def projects(self):
        """returns all the projects in the studio
        """
        from stalker import Project
        return Project.query.all()

    @property
    def active_projects(self):
        """returns all the active projects in the studio
        """
        from stalker import Project
        return Project.query.filter_by(active=True).all()

    @property
    def inactive_projects(self):
        """return all the inactive projects in the studio
        """
        from stalker import Project
        return Project.query.filter_by(active=False).all()

    @property
    def departments(self):
        """returns all the departments in the studio
        """
        from stalker import Department
        return Department.query.all()

    @property
    def users(self):
        """returns all the users in the studio
        """
        from stalker import User
        return User.query.all()

    @property
    def vacations(self):
        """returns all Vacations which doesn't have a User defined
        """
        return Vacation.query.filter(Vacation.user==None).all()

    def schedule(self, scheduled_by=None):
        """Schedules all the active projects in the studio. Needs a Scheduler,
        so before calling it set a scheduler by using the :attr:`.scheduler`
        attribute.

        :param scheduled_by: A User instance who is doing the scheduling.
        """
        # check the scheduler first
        if self.scheduler is None or \
                not isinstance(self.scheduler, SchedulerBase):
            raise RuntimeError(
                'There is no scheduler for this %(class)s, please assign a '
                'scheduler to the %(class)s.scheduler attribute, before '
                'calling %(class)s.schedule()' %
                {
                    'class': self.__class__.__name__
                }
            )

        with db.DBSession.no_autoflush:
            self.scheduling_started_at = datetime.datetime.now()

            # run the scheduler
            self.scheduler.studio = self
        start = time.time()

        # commit before scheduling
        #DBSession.commit()

        result = None
        try:
            result = self.scheduler.schedule()
        finally:
            # in any case set is_scheduling to False
            with db.DBSession.no_autoflush:
                self.is_scheduling = False
                self.is_scheduling_by = None

                # also store the result
                # if result:
                self.last_schedule_message = result

                # And the date the schedule is completed
                # TODO: convert to UTC time
                self.last_scheduled_at = datetime.datetime.now()

                # and who has done the scheduling
                if scheduled_by:
                    logger.debug(
                        'setting last_scheduled_by to : %s' % scheduled_by
                    )
                    self.last_scheduled_by = scheduled_by

        end = time.time()
        logger.debug('scheduling took %s seconds' % (end - start))
        return result

    @property
    def weekly_working_hours(self):
        """returns the WorkingHours.weekly_working_hours
        """
        return self.working_hours.weekly_working_hours

    @property
    def weekly_working_days(self):
        """returns the WorkingHours.weekly_working_hours
        """
        return self.working_hours.weekly_working_days

    @property
    def yearly_working_days(self):
        """returns the yearly working days
        """
        return self.working_hours.yearly_working_days

    def to_unit(self, from_timing, from_unit, to_unit, working_hours=True):
        """converts the given timing and unit to the desired unit
        if working_hours=True then the given timing is considered as working
        hours
        """
        raise NotImplementedError('this is not implemented yet')

    def _timing_resolution_getter(self):
        """returns the timing_resolution
        """
        return self._timing_resolution

    def _timing_resolution_setter(self, res_in):
        """sets the timing_resolution
        """
        self._timing_resolution = self._validate_timing_resolution(res_in)
        logger.debug('self._timing_resolution: %s' % self._timing_resolution)
        # update date values
        if self.start and self.end and self.duration:
            self._start, self._end, self._duration = \
                self._validate_dates(
                    self.round_time(self.start),
                    self.round_time(self.end),
                    None
                )

    timing_resolution = synonym(
        '_timing_resolution',
        descriptor=property(
            _timing_resolution_getter,
            _timing_resolution_setter,
            doc="""The timing_resolution of this object.

            Can be set to any value that is representable with
            datetime.timedelta. The default value is 1 hour. Whenever it is
            changed the start, end and duration values will be updated.
            """
        )
    )

    def _validate_timing_resolution(self, timing_resolution):
        """validates the given timing_resolution value
        """
        if timing_resolution is None:
            timing_resolution = defaults.timing_resolution

        if not isinstance(timing_resolution, datetime.timedelta):
            raise TypeError(
                '%s.timing_resolution should be an instance of '
                'datetime.timedelta not, %s' %
                (self.__class__.__name__, timing_resolution.__class__.__name__)
            )

        return timing_resolution


class WorkingHours(object):
    """A helper class to manage Studio working hours.

    Working hours is a data class to store the weekly working hours pattern of
    the studio.

    The data stored as a dictionary with the short day names are used as the
    key and the value is a list of two integers showing the working hours
    interval as the minutes after midnight. This is done in that way to ease
    the data transfer to TaskJuggler. The default value is defined in
    :class:`stalker.config.Config` ::

      wh = WorkingHours()
      wh.working_hours = {
          'mon': [[540, 720], [820, 1080]], # 9:00 - 12:00, 13:00 - 18:00
          'tue': [[540, 720], [820, 1080]], # 9:00 - 12:00, 13:00 - 18:00
          'wed': [[540, 720], [820, 1080]], # 9:00 - 12:00, 13:00 - 18:00
          'thu': [[540, 720], [820, 1080]], # 9:00 - 12:00, 13:00 - 18:00
          'fri': [[540, 720], [820, 1080]], # 9:00 - 12:00, 13:00 - 18:00
          'sat': [], # saturday off
          'sun': [], # sunday off
      }

    The default value is 9:00 - 18:00 from Monday to Friday and Saturday and
    Sunday are off.

    The working hours can be updated by the user supplied dictionary. If the
    user supplied dictionary doesn't have all the days then the default values
    will be used for those days.

    It is possible to use day index and day short names as a key value to reach
    the data::

      from stalker import config
      defaults = config.Config()

      wh = WorkingHours()

      # this is same by doing wh.working_hours['sun']
      assert wh['sun'] == defaults.working_hours['sun']

      # you can reach the data using the weekday number as index
      assert wh[0] == defaults.working_hours['mon']

      # working hours of sunday if defaults are used or any other day defined
      # by the stalker.config.Config.day_order
      assert wh[0] == defaults.working_hours[defaults.day_order[0]]

    :param working_hours: The dictionary that shows the working hours. The keys
      of the dictionary should be one of ['mon', 'tue', 'wed', 'thu', 'fri',
      'sat', 'sun']. And the values should be a list of two integers like
      [[int, int], [int, int], ...] format, showing the minutes after midnight.
      For missing days the default value will be used. If skipped the default
      value is going to be used.
    """

    def __init__(self,
                 working_hours=None,
                 daily_working_hours=None,
                 **kwargs):
        if working_hours is None:
            working_hours = defaults.working_hours
        self._wh = None
        self.working_hours = self._validate_working_hours(working_hours)
        self._daily_working_hours = None
        self.daily_working_hours = daily_working_hours

    def __eq__(self, other):
        """overridden equality operator
        """
        return isinstance(other, WorkingHours) and \
            other.working_hours == self.working_hours

    def __hash__(self):
        """the overridden __hash__ method
        """
        return hash(self.working_hours)

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._wh[defaults.day_order[item]]
        elif isinstance(item, str):
            return self._wh[item]

    def __setitem__(self, key, value):
        self._validate_wh_value(value)
        if isinstance(key, int):
            self._wh[defaults.day_order[key]] = value
        elif isinstance(key, str):
            # check if key is in
            if key not in defaults.day_order:
                raise KeyError(
                    '%s accepts only %s as key, not %s' %
                    (self.__class__.__name__, defaults.day_order, key)
                )
            self._wh[key] = value

    def _validate_working_hours(self, wh_in):
        """validates the given working hours
        """
        if not isinstance(wh_in, dict):
            raise TypeError(
                '%s.working_hours should be a dictionary, not %s' %
                (self.__class__.__name__, wh_in.__class__.__name__)
            )

        for key in wh_in.keys():
            if not isinstance(wh_in[key], list):
                raise TypeError(
                    '%s.working_hours should be a dictionary with keys "mon, '
                    'tue, wed, thu, fri, sat, sun" and the values should a '
                    'list of lists of two integers like [[540, 720], [800, '
                    '1080]], not %s' %
                    (self.__class__.__name__, wh_in[key].__class__.__name__)
                )

            # validate item values
            self._validate_wh_value(wh_in[key])

        # update the default values with the supplied working_hour dictionary
        # copy the defaults
        wh_def = copy.copy(defaults.working_hours)
        # update them
        wh_def.update(wh_in)

        return wh_def

    @property
    def working_hours(self):
        """the getter of _wh
        """
        return self._wh

    @working_hours.setter
    def working_hours(self, wh_in):
        """the setter of _wh
        """
        self._wh = self._validate_working_hours(wh_in)

    def is_working_hour(self, check_for_date):
        """checks if the given datetime is in working hours

        :param datetime.datetime check_for_date: The time to check if it is a
          working hour
        """
        weekday_nr = check_for_date.weekday()
        hour = check_for_date.hour
        minute = check_for_date.minute

        time_from_midnight = hour * 60 + minute

        # check if the hour is inside the working hour ranges
        logger.debug('checking for: %s' % time_from_midnight)
        logger.debug('self[weekday_nr]: %s' % self[weekday_nr])
        for working_hour_groups in self[weekday_nr]:
            start = working_hour_groups[0]
            end = working_hour_groups[1]
            logger.debug('start       : %s' % start)
            logger.debug('end         : %s' % end)
            if start <= time_from_midnight < end:
                return True

        return False

    def _validate_wh_value(self, value):
        """validates the working hour value
        """
        err = '%s.working_hours value should be a list of lists of two ' \
              'integers between and the range of integers should be 0-1440, ' \
              'not %s'

        if not isinstance(value, list):
            raise TypeError(err % (self.__class__.__name__,
                                   value.__class__.__name__))

        for i in value:
            if not isinstance(i, list):
                raise TypeError(err % (self.__class__.__name__,
                                       i.__class__.__name__))

            # check list length
            if len(i) != 2:
                raise RuntimeError(err % (self.__class__.__name__, value))

            # check type
            if not isinstance(i[0], int) or not isinstance(i[1], int):
                raise TypeError(err % (self.__class__.__name__, value))

            # check range
            if i[0] < 0 or i[0] > 1440 or i[1] < 0 or i[1] > 1440:
                raise ValueError(err % (self.__class__.__name__, value))

        return value

    @property
    def to_tjp(self):
        """returns TaskJuggler representation of this object
        """
        # render the template
        from jinja2 import Template

        template = Template(defaults.tjp_working_hours_template)
        return template.render({'workinghours': self})

    @property
    def weekly_working_hours(self):
        """returns the total working hours in a week
        """
        weekly_working_hours = 0
        for i in range(0, 7):
            for start, end in self[i]:
                weekly_working_hours += (end - start)
        return weekly_working_hours / 60.0

    @property
    def weekly_working_days(self):
        """returns the weekly working days by looking at the working hours
        settings
        """
        wwd = 0
        for i in range(0, 7):
            if len(self[i]):
                wwd += 1
        return wwd

    @property
    def yearly_working_days(self):
        """returns the total working days in a year
        """
        return int(ceil(self.weekly_working_days * 52.1428))

    def _validate_daily_working_hours(self, dwh):
        """validates the given daily working hours value
        """
        if dwh is None:
            dwh = defaults.daily_working_hours

        if not isinstance(dwh, int):
            raise TypeError(
                '%s.daily_working_hours should be an integer, not %s' %
                (self.__class__.__name__, dwh.__class__.__name__)
            )

        if dwh <= 0 or dwh > 24:
            raise ValueError(
                '%s.daily_working_hours should be a positive integer value '
                'greater than 0 and smaller than or equal to 24'
            )
        return dwh

    @property
    def daily_working_hours(self):
        """getter for daily_working_hours attribute
        """
        return self._daily_working_hours

    @daily_working_hours.setter
    def daily_working_hours(self, dwh):
        """setter for daily_working_hours attribute
        """
        self._daily_working_hours = self._validate_daily_working_hours(dwh)

    def split_in_to_working_hours(self, start, end):
        """splits the given start and end datetime objects in to working hours
        """
        raise NotImplementedError()


class Vacation(SimpleEntity, DateRangeMixin):
    """Vacation is the way to manage the User vacations.

    :param user: The user of this vacation. Should be an instance of
      :class:`.User` if skipped or given as None the
      Vacation is considered as a Studio vacation and applies to all Users.

    :param start: The start datetime of the vacation. Is is an
      datetime.datetime instance. When skipped it will be set to the rounded
      value of.

    :param end: The end datetime of the vacation. It is an datetime.datetime
      instance.
    """
    __auto_name__ = True
    __tablename__ = 'Vacations'
    __mapper_args__ = {'polymorphic_identity': 'Vacation'}

    __strictly_typed__ = False

    vacation_id = Column("id", Integer, ForeignKey("SimpleEntities.id"),
                         primary_key=True)

    user_id = Column('user_id', Integer, ForeignKey('Users.id'),
                     nullable=True)

    user = relationship(
        'User',
        primaryjoin='Vacations.c.user_id==Users.c.id',
        back_populates='vacations',
        doc="""The User of this Vacation.

        Accepts :class:`.User` instance.
        """
    )

    def __init__(self, user=None, start=None, end=None, **kwargs):
        kwargs['start'] = start
        kwargs['end'] = end
        super(Vacation, self).__init__(**kwargs)
        DateRangeMixin.__init__(self, **kwargs)
        self.user = user

    @validates('user')
    def _validate_user(self, key, user):
        """validates the given user instance
        """
        if user is not None:
            from stalker import User
            if not isinstance(user, User):
                raise TypeError(
                    '%s.user should be an instance of '
                    'stalker.models.auth.User, not %s' %
                    (self.__class__.__name__, user.__class__.__name__)
                )
        return user

    @property
    def to_tjp(self):
        """overridden to_tjp method
        """
        from jinja2 import Template

        template = Template(defaults.tjp_vacation_template)
        return template.render({'vacation': self})



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2014 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os
import logging

from sqlalchemy import Column, Integer, ForeignKey, String, Text
from sqlalchemy.orm import validates

from stalker.models.entity import Entity
from stalker.log import logging_level

logger = logging.getLogger(__name__)
logger.setLevel(logging_level)


class Link(Entity):
    """Holds data about external links.

    Links are all about giving some external information to the current entity
    (external to the database, so it can be something on the
    :class:`.Repository` or in the Web or anywhere that the server can reach).
    The type of the link (general, file, folder, web page, image, image
    sequence, video, movie, sound, text etc.) can be defined by a
    :class:`.Type` instance (you can also use multiple :class:`.Tag` instances
    to add more information, and to filter them back). Again it is defined by
    the needs of the studio.

    For sequences of files the file name should be in "%h%p%t %R" format in
    PySeq_ formatting rules.

    There are three secondary attributes (properties to be more precise)
    ``path``, ``filename`` and ``extension``. These attributes are derived from
    the :attr:`.full_path` attribute and they modify it.

    Path
      It is the path part of the full_path

    Filename
      It is the filename part of the full_path, also includes the extension,
      so changing the filename also changes the extension part.

    Extension
      It is the extension part of the full_path. It also includes the extension
      separator ('.' for most of the file systems).

    :param full_path: The full path to the link, it can be a path to a folder
      or a file in the file system, or a web page. For file sequences use
      "%h%p%t %R" format, for more information see `PySeq Documentation`_.
      It can be set to empty string (or None which will be converted to an
      empty string automatically).

    .. _PySeq: http://packages.python.org/pyseq/
    .. _PySeq Documentation: http://packages.python.org/pyseq/

    """
    __auto_name__ = True
    __tablename__ = "Links"
    __mapper_args__ = {"polymorphic_identity": "Link"}

    link_id = Column(
        "id",
        Integer,
        ForeignKey("Entities.id"),
        primary_key=True,
    )

    original_filename = Column(String(256))  # this is a limit for most
                                             # filesystems
    full_path = Column(
        Text,
        doc="""The full path of the url to the link."""
    )

    def __init__(self, full_path='', original_filename='', **kwargs):
        super(Link, self).__init__(**kwargs)
        self.full_path = full_path
        self.original_filename = original_filename

    @validates('full_path')
    def _validate_full_path(self, key, full_path):
        """validates the given full_path value
        """
        if full_path is None:
            full_path = ''

        if not isinstance(full_path, str):
            raise TypeError(
                "%s.full_path should be an instance of string not %s" %
                (self.__class__.__name__, full_path.__class__.__name__)
            )

        return self._format_path(full_path)

    @validates('original_filename')
    def _validate_original_filename(self, key, original_filename):
        """validates the given original_filename value
        """
        filename_from_path = os.path.basename(self.full_path)
        if original_filename is None:
            original_filename = filename_from_path

        if original_filename == '':
            original_filename = filename_from_path

        if not isinstance(original_filename, str):
            raise TypeError(
                '%s.original_filename should be an instance of str and not '
                '%s' % (
                    self.__class__.__name__,
                    original_filename.__class__.__name__
                )
            )

        return original_filename

    @staticmethod
    def _format_path(path):
        """formats the path to internal format, which is Linux forward slashes
        for path separation
        """
        if not isinstance(path, str):
            path = path.encode('utf-8')

        return path.replace("\\", "/")

    @property
    def path(self):
        """the path property
        """
        return os.path.split(self.full_path)[0]

    @path.setter
    def path(self, path):
        """setter for the path

        :param str path: the new path
        """
        if path is None:
            raise TypeError(
                '%s.path can not be set to None' % self.__class__.__name__
            )

        if not isinstance(path, str):
            raise TypeError(
                '%s.path should be an instance of str, not %s' %
                (self.__class__.__name__, path.__class__.__name__)
            )

        if path == '':
            raise ValueError(
                '%s.path can not be an empty string' % self.__class__.__name__
            )

        self.full_path = self._format_path(
            os.path.join(path, self.filename)
        )

    @property
    def filename(self):
        """the filename property
        """
        return os.path.split(self.full_path)[1]

    @filename.setter
    def filename(self, filename):
        """filename setter

        :param str filename: the new filename
        """
        if filename is None:
            filename = ''

        if not isinstance(filename, str):
            raise TypeError(
                '%s.filename should be an instance of str, not %s' %
                (self.__class__.__name__, filename.__class__.__name__)
            )

        self.full_path = self._format_path(
            os.path.join(self.path, filename)
        )

    @property
    def extension(self):
        """the extension property
        """
        return os.path.splitext(self.full_path)[1]

    @extension.setter
    def extension(self, extension):
        """extension setter

        :param extension: the new extension
        """
        if extension is None:
            extension = ''

        if not isinstance(extension, str):
            raise TypeError(
                '%s.extension should be an instance of str, not %s' % (
                    self.__class__.__name__,
                    extension.__class__.__name__
                )
            )

        if extension != '':
            if not extension.startswith(os.path.extsep):
                extension = os.path.extsep + extension

        self.filename = os.path.splitext(self.filename)[0] + extension

    def __eq__(self, other):
        """the equality operator
        """
        return super(Link, self).__eq__(other) and \
            isinstance(other, Link) and \
            self.full_path == other.full_path and \
            self.type == other.type

    def __hash__(self):
        """the overridden __hash__ method
        """
        return hash(self.id) + 2 * hash(self.name) + 3 * hash(self.entity_type)



import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README')).read()
CHANGES = open(os.path.join(here, 'CHANGELOG')).read()

requires = [
    'sqlalchemy>=0.8',
    'alembic',
    'jinja2',
]

setup(
    name='stalker',
    version='0.2.8.1.1',
    description='A Production Asset Management (ProdAM) System',
    long_description=README,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 3.0",
        "License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "Topic :: Database",
        "Topic :: Software Development",
        "Topic :: Utilities",
        "Topic :: Office/Business :: Scheduling",
    ],
    author='Erkan Ozgur Yilmaz',
    author_email='eoyilmaz@gmail.com',
    url='http://github.com/eoyilmaz/stalker',
    keywords=['production', 'asset', 'management', 'vfx', 'animation',
              'maya', 'houdini', 'nuke', 'fusion', 'softimage', 'blender',
              'vue'],
    packages=find_packages(),
    include_package_data=True,
    data_files=[
        ('', [
            'COPYING',
            'COPYING.LESSER',
            'INSTALL',
            'MANIFEST.in',
            'README'
        ]),
    ],
    zip_safe=True,
    test_suite='stalker',
    install_requires=requires,
)

